from __future__ import annotations

import functools as ft

import jax
import jax.numpy as jnp


def _check_and_return(value, ndim, name):
    if isinstance(value, int):
        return (value,) * ndim
    elif isinstance(value, jax.Array):
        return jnp.repeat(value, ndim)
    elif isinstance(value, tuple):
        assert len(value) == ndim, f"{name} must be a tuple of length {ndim}"
        return tuple(value)
    raise ValueError(f"Expected int or tuple for {name}, got {value}.")


def _generate_forward_offsets(
    derivative: int, accuracy: int = 2
) -> tuple[float | int, ...]:
    """Generate central difference offsets

    Args:
        derivative : derivative order

    Returns:
        tuple[float | int, ...]: central difference offsets
    """
    if derivative < 1:
        msg = f"derivative must be >= 1 for forward offset generation, got {derivative}"
        raise ValueError(msg)
    if accuracy < 1:
        msg = f"accuracy must be >= 2 for forward offset generation, got {accuracy}"
        raise ValueError(msg)

    return tuple(range(0, (derivative + accuracy)))


def _generate_central_offsets(
    derivative: int, accuracy: int = 2
) -> tuple[float | int, ...]:
    """Generate central difference offsets

    Args:
        derivative : derivative order

    Returns:
        tuple[float | int, ...]: central difference offsets
    """
    if derivative < 1:
        msg = f"derivative must be >= 1 for central offset generation, got {derivative}"
        raise ValueError(msg)
    if accuracy < 2:
        msg = f"accuracy must be >= 2 for central offset generation, got {accuracy}"
        raise ValueError(msg)

    return tuple(
        range(-((derivative + accuracy - 1) // 2), (derivative + accuracy - 1) // 2 + 1)
    )


def _generate_backward_offsets(
    derivative: int, accuracy: int = 2
) -> tuple[float | int, ...]:
    """Generate central difference offsets

    Args:
        derivative : derivative order

    Returns:
        tuple[float | int, ...]: central difference offsets
    """
    if derivative < 1:
        msg = f"derivative must be >= 1 for back offset generation, got {derivative}"
        raise ValueError(msg)
    if accuracy < 1:
        msg = f"accuracy must be >= 2 for back offset generation, got {accuracy}"
        raise ValueError(msg)

    return tuple(range(-(derivative + accuracy - 1), 1))


@ft.partial(jax.jit, static_argnums=(1,))
def generate_finitediff_coeffs(
    offsets: tuple[float | int, ...], derivative: int
) -> tuple[float]:
    """Generate FD coeffs

    Args:
        offsets: offsets of the finite difference stencil
        derivative: derivative order

    Returns:
        tuple[float]: finite difference coefficients

    Example:
        >>> generate_finitediff_coeffs(offsets=(-1, 0, 1), derivative=1)
        (-0.5, 0.0, 0.5)

        >>> generate_finitediff_coeffs(offsets=(-1, 0, 1), derivative=2)
        (1.0, -2.0, 1.0)
        # translates to  1*f(x-1) - 2*f(x) + 1*f(x+1)

    See:
        https://en.wikipedia.org/wiki/Finite_difference_coefficient
        https://web.media.mit.edu/~crtaylor/calculator.html
    """

    N = len(offsets)

    if derivative >= N:
        msg = "Sampling points must be larger than derivative order."
        msg += f" len(offsets)={len(offsets)} must be less than {derivative}"
        raise ValueError(msg)

    A = jnp.repeat(jnp.array(offsets)[None, :], repeats=N, axis=0)
    A **= jnp.arange(0, N).reshape(-1, 1)
    index = jnp.arange(N)
    factorial = jnp.prod(jnp.arange(1, derivative + 1))
    B = jnp.where(index == derivative, factorial, 0)[:, None]
    C = jnp.linalg.inv(A) @ B  # solve Ax = B
    return C.flatten()



import jax
import jax.numpy as jnp
import numpy as np
import pytest
from jax.experimental import enable_x64

from finitediffx import fgrad, generate_finitediff_coeffs


def test_generate_finitediff_coeffs():
    with enable_x64():
        DF = lambda N: generate_finitediff_coeffs(N, 1)
        all_correct = lambda x, y: np.testing.assert_allclose(x, y, atol=1e-2)

        # https://en.wikipedia.org/wiki/fgraderence_coefficient
        all_correct(DF((0, 1)), jnp.array([-1.0, 1.0]))
        all_correct(DF((0, 1, 2)), jnp.array([-3 / 2.0, 2.0, -1 / 2]))
        all_correct(DF((0, 1, 2, 3)), jnp.array([-11 / 6, 3.0, -3 / 2, 1 / 3]))
        all_correct(
            DF((0, 1, 2, 3, 4)), jnp.array([-25 / 12, 4.0, -3.0, 4 / 3, -1 / 4])
        )
        all_correct(
            DF((0, 1, 2, 3, 4, 5)),
            jnp.array([-137 / 60, 5.0, -5, 10 / 3, -5 / 4, 1 / 5]),
        )
        all_correct(
            DF((0, 1, 2, 3, 4, 5, 6)),
            jnp.array([-49 / 20, 6.0, -15 / 2, 20 / 3, -15 / 4, 6 / 5, -1 / 6]),
        )

        # https://web.media.mit.edu/~crtaylor/calculator.html
        all_correct(DF((-2.2, 3.2, 5)), jnp.array([-205 / 972, 280 / 972, -75 / 972]))

        # https://web.njit.edu/~jiang/math712/fornberg.pdf  Table 4.
        DF = lambda N: generate_finitediff_coeffs(N, 0)
        all_correct(DF((-0.5,)), jnp.array([1]))
        all_correct(DF((-0.5, 0.5)), jnp.array([0.5, 0.5]))
        all_correct(DF((-0.5, 0.5, 1.5)), jnp.array([3 / 8, 3 / 4, -1 / 8]))
        all_correct(
            DF((-0.5, 0.5, 1.5, 2.5)), jnp.array([5 / 16, 15 / 16, -5 / 16, 1 / 16])
        )


def test_fgrad_args():
    with enable_x64():
        all_correct = lambda lhs, rhs: np.testing.assert_allclose(lhs, rhs, atol=0.05)

        for func in [
            lambda x: x,
            lambda x: x**2,
            lambda x: x + 2,
            lambda x: jnp.sin(x),
            lambda x: jnp.cos(x) * jnp.sin(x),
        ]:
            f1 = jax.grad(func)
            f2 = fgrad(func)
            args = (1.5,)
            F1, F2 = f1(*args), f2(*args)
            all_correct(F1, F2)

        for func in [
            lambda x, y: x + y,
            lambda x, y: x**2 + y**3,
            lambda x, y: x + 2 + y * x,
        ]:
            f1 = jax.grad(func)
            f2 = fgrad(func)
            args = (1.5, 2.5)
            F1, F2 = f1(*args), f2(*args)
            all_correct(F1, F2)

        for func in [
            lambda x, y, z: x + y + z,
            lambda x, y, z: x**2 + y**3 * z,
            lambda x, y, z: x + 2 + y * x + z,
        ]:
            f1 = jax.grad(func)
            f2 = fgrad(func)
            args = (1.5, 2.5, 3.5)
            F1, F2 = f1(*args), f2(*args)
            all_correct(F1, F2)


def test_fgrad_second_derivative():
    with enable_x64():
        all_correct = lambda lhs, rhs: np.testing.assert_allclose(lhs, rhs, atol=0.05)

        for func in [lambda x: x, lambda x: x**2, lambda x: x + 2]:
            f1 = jax.grad(jax.grad(func))
            f2 = fgrad(func, derivative=2)
            f3 = fgrad(fgrad(func))
            args = (1.5,)
            F1, F2, F3 = f1(*args), f2(*args), f3(*args)
            all_correct(F1, F2)
            all_correct(F1, F3)

        for func in [
            lambda x, y: x + y,
            lambda x, y: x**2 + y**3,
            lambda x, y: x + 2 + y * x,
        ]:
            f1 = jax.grad(jax.grad(func))
            f2 = fgrad(func, derivative=2)
            f3 = fgrad(fgrad(func))
            args = (1.5, 2.5)
            F1, F2, F3 = f1(*args), f2(*args), f3(*args)
            all_correct(F1, F2)
            all_correct(F1, F3)

        for func in [
            lambda x, y, z: x + y + z,
            lambda x, y, z: x**2 + y**3 * z,
            lambda x, y, z: x + 2 + y * x + z,
        ]:
            f1 = jax.grad(jax.grad(func))
            f2 = fgrad(func, derivative=2)
            f3 = fgrad(fgrad(func))
            args = (1.5, 2.5, 3.5)
            F1, F2, F3 = f1(*args), f2(*args), f3(*args)
            all_correct(F1, F2)
            all_correct(F1, F3)


def test_fgrad_argnum():
    with enable_x64():
        all_correct = lambda lhs, rhs: np.testing.assert_allclose(lhs, rhs, atol=0.05)

        # test argnums
        func = lambda x, y, z: x**2 + y**3 + z**4
        f1 = jax.grad(func, argnums=(0,))(1.0, 1.0, 1.0)
        f2 = fgrad(func, argnums=0)(1.0, 1.0, 1.0)
        all_correct(f1, f2)

        f1 = jax.grad(func, argnums=(1,))(1.0, 1.0, 1.0)
        f2 = fgrad(func, argnums=1)(1.0, 1.0, 1.0)
        all_correct(f1, f2)

        f1 = jax.grad(func, argnums=(2,))(1.0, 1.0, 1.0)
        f2 = fgrad(func, argnums=2)(1.0, 1.0, 1.0)
        all_correct(f1, f2)

        # multiple argnums
        f1l, f1r = jax.grad(func, argnums=(0, 1))(1.0, 1.0, 1.0)
        f2l, f2r = fgrad(func, argnums=(0, 1))(1.0, 1.0, 1.0)
        all_correct(f1l, f2l)
        all_correct(f1r, f2r)


def test_fgrad_error():
    with pytest.raises(ValueError):
        fgrad(lambda x: x, argnums=-1)

    with pytest.raises(ValueError):
        fgrad(lambda x: x, argnums=1.0)

    with pytest.raises(ValueError):
        fgrad(lambda x: x, accuracy=1)

    with pytest.raises(ValueError):
        fgrad(lambda x: x, argnums=[1, 2])



# credits to Mahmoud Asem 2022 @KAIST

# This file defines the fgrad function which is a finite difference approximation of jax.grad
# for first order derivative, finite difference and automatic difference is approximated as the following
# f(x + Δx) = f(x) + f'(x) Δx + f''(x) Δx^2 / 2 + f'''(x) Δx^3 / 6 + ...
# f(x + ɛ Δx) = f(x) + f'(x) ɛ Δx + f''(x) ɛ^2 Δx^2 / 2 + f'''(x) ɛ^3 Δx^3 / 6 + ... , where ɛ^2 := 0 and ɛ != 0


from __future__ import annotations

import functools as ft
from typing import Callable

import jax
import jax.numpy as jnp

from finitediffx._src.utils import _generate_central_offsets, generate_finitediff_coeffs


def _evaluate_func_at_shifted_steps_along_argnum(
    func: Callable,
    *,
    coeffs: jax.Array,
    offsets: tuple[float | int, ...],
    argnum: int,
    step_size: float,
    derivative: int,
):
    if not isinstance(argnum, int) or argnum < 0:
        raise ValueError(f"argnum must be a non-negative integer, got {argnum}")

    DX = jnp.array(offsets) * step_size

    def wrapper(*args, **kwargs):
        # yield function output at shifted points
        for coeff, dx in zip(coeffs, DX):
            shifted_args = list(args)
            shifted_args[argnum] += dx
            yield coeff * func(*shifted_args, **kwargs) / (step_size**derivative)

    return wrapper


def fgrad(
    func: Callable,
    *,
    argnums: int = 0,
    step_size: float = None,
    offsets: tuple[float | int, ...] = None,
    derivative: int = 1,
    accuracy: int = 3,
) -> Callable:
    """Finite difference derivative of a function with respect to one of its arguments.
    similar to jax.grad but with finite difference approximation

    This function could be useful in certain situations for example
    >>> import jax
    >>> jax.config.update("jax_enable_x64", True)
    >>> f = lambda x : x**10
    >>> def repeated_grad(f, n):
    ...     return f if n==0 else jax.grad(repeated_grad(f, n-1))

    >>> df_grad = jax.jit(repeated_grad(f,10))
    >>> %timeit df_grad(1.)
    4.83 µs ± 4.3 µs per loop (mean ± std. dev. of 7 runs, 1 loop each)

    >>> df_fgrad = jax.jit(fgrad(f, argnums=0, derivative=10, accuracy=2))
    >>> %timeit df_fgrad(1.)
    2.45 µs ± 63.7 ns per loop (mean ± std. dev. of 7 runs, 100,000 loops each)

    Args:
        func: function to differentiate
        argnums:
            argument number to differentiate. Defaults to 0.
            If a tuple is passed, the function is differentiated with respect to all the arguments in the tuple.
        step_size: step size for the finite difference stencil. Defaults to None.
        offsets: offsets for the finite difference stencil. Defaults to None.
        derivative: derivative order. Defaults to 1.
        accuracy: accuracy of the finite difference stencil. Defaults to 2. used to generate offsets if not provided.

    Returns:
        Callable: derivative of the function

    Example:
        >>> def f(x):
        ...     return x**2
        >>> df = fgrad(f)
        >>> df(2.0)
        DeviceArray(4., dtype=float32)
    """
    if offsets is None:
        if accuracy < 2:
            raise ValueError(f"accuracy must be >= 2, got {accuracy}")
        # generate central offsets based on accuracy
        offsets = _generate_central_offsets(derivative, accuracy=accuracy)

    if step_size is None:
        # the best default step size for arbitrary derivative order and accuracy
        step_size = (2) ** (-23 / (2 * derivative))

    # finite difference coefficients
    coeffs = generate_finitediff_coeffs(offsets, derivative)

    # TODO: edit docstring of the differentiated function

    if isinstance(argnums, int):
        dfunc = _evaluate_func_at_shifted_steps_along_argnum(
            func,
            coeffs=coeffs,
            offsets=offsets,
            argnum=argnums,
            step_size=step_size,
            derivative=derivative,
        )
        return ft.wraps(func)(lambda *a, **k: sum(dfunc(*a, **k)))

    if isinstance(argnums, tuple):
        # return a tuple of derivatives if argnums is a tuple
        dfuncs = [
            _evaluate_func_at_shifted_steps_along_argnum(
                func,
                coeffs=coeffs,
                offsets=offsets,
                argnum=argnum,
                step_size=step_size,
                derivative=derivative,
            )
            for argnum in argnums
        ]
        return ft.wraps(func)(lambda *a, **k: tuple(sum(df(*a, **k)) for df in dfuncs))

    raise ValueError(f"argnums must be an int or a tuple of ints, got {argnums}")



from ._src.fgrad import fgrad
from ._src.finite_diff import (
    curl,
    difference,
    divergence,
    gradient,
    hessian,
    jacobian,
    laplacian,
)
from ._src.utils import generate_finitediff_coeffs

__all__ = (
    "curl",
    "divergence",
    "difference",
    "gradient",
    "jacobian",
    "laplacian",
    "hessian",
    "fgrad",
    "generate_finitediff_coeffs",
)


__version__ = "0.0.2"


# import jax
import jax.numpy as jnp
import numpy.testing as npt
import pytest
from jax.experimental import enable_x64

from finitediffx import (
    curl,
    difference,
    divergence,
    gradient,
    hessian,
    jacobian,
    laplacian,
)
from finitediffx._src.utils import (
    _generate_backward_offsets,
    _generate_central_offsets,
    _generate_forward_offsets,
    generate_finitediff_coeffs,
)


def test_difference_central():
    with enable_x64():
        x, y = [jnp.linspace(0, 1, 1000)] * 2
        dx, dy = x[1] - x[0], y[1] - y[0]
        X, Y = jnp.meshgrid(x, y, indexing="ij")
        F = jnp.sin(X) * jnp.cos(Y)
        dFdX = difference(F, step_size=dx, axis=0, accuracy=3)
        npt.assert_allclose(dFdX, jnp.cos(X) * jnp.cos(Y), rtol=1e-3)

        dFdY = difference(F, step_size=dy, axis=1, accuracy=3)
        npt.assert_allclose(dFdY, -jnp.sin(X) * jnp.sin(Y), atol=1e-7)

        # 1 4 6 3 2
        # left = (4-1) = 3 -> forward difference
        # center = (6-1)/2, (3-4)/2 -> central difference
        # right = (2-3) = -1 -> backward difference
        npt.assert_allclose(
            difference(jnp.array([1, 4, 6, 3, 2]), accuracy=1, axis=0, derivative=1),
            jnp.array([3, 2.5, -0.5, -2, -1]),
        )

        # 1 4 6 3 2 4
        # [-1.5  2.  -0.5] (0, 1, 2)
        # [-0.5  0.   0.5] (-1, 0, 1)
        # [ 0.5 -2.   1.5] (-2, -1, 0)
        npt.assert_allclose(
            difference(jnp.array([1, 4, 6, 3, 2, 4]), accuracy=2, axis=0, derivative=1),
            jnp.array([3.5, 2.5, -0.5, -2.0, 0.5, 3.5]),
        )

        # no central difference possible
        x = jnp.array([1.2, 1.4])
        npt.assert_allclose(
            difference(x, accuracy=1, axis=0, derivative=1), jnp.array([0.2, 0.2])
        )


def test_difference_forward():
    with enable_x64():
        x, y = [jnp.linspace(0, 1, 1000)] * 2
        dx, dy = x[1] - x[0], y[1] - y[0]
        X, Y = jnp.meshgrid(x, y, indexing="ij")
        F = jnp.sin(X) * jnp.cos(Y)
        dFdX = difference(F, step_size=dx, axis=0, accuracy=3)
        npt.assert_allclose(dFdX, jnp.cos(X) * jnp.cos(Y), rtol=1e-3)

        dFdY = difference(F, step_size=dy, axis=1, accuracy=3)
        npt.assert_allclose(dFdY, -jnp.sin(X) * jnp.sin(Y), atol=1e-7)

        # 1 4 6 3 2
        # left = (4-1), (6-4), (3-6), (2-3)-> forward difference
        # right = (2-3) = -1 -> backward difference
        npt.assert_allclose(
            difference(jnp.array([1, 4, 6, 3, 2]), accuracy=1, axis=0, derivative=1, method="forward"),
            jnp.array([3., 2., -3., -1., -1.]),
        )

        # 1 4 6 3 2 4
        # Forward Coeffs -> [-1.5  2.  -0.5] (0, 1, 2)
        # Backward Coeffs -> [ 0.5 -2.   1.5] (-2, -1, 0)
        npt.assert_allclose(
            difference(jnp.array([1, 4, 6, 3, 2, 4]), accuracy=2, axis=0, derivative=1, method="forward"),
            jnp.array([3.5, 4.5, -4.0, -2.5, 0.0, 3.5]),
        )

        # no central difference possible
        x = jnp.array([1.2, 1.4])
        # Forward Coeffs -> [-1 1] (0, 1)
        # Backward Coeffs -> [1 -1] (-1, 0)
        npt.assert_allclose(
            difference(x, accuracy=1, axis=0, derivative=1, method="forward"), jnp.array([0.2, 0.2])
        )


def test_difference_backward():
    with enable_x64():
        x, y = [jnp.linspace(0, 1, 1000)] * 2
        dx, dy = x[1] - x[0], y[1] - y[0]
        X, Y = jnp.meshgrid(x, y, indexing="ij")
        F = jnp.sin(X) * jnp.cos(Y)
        dFdX = difference(F, step_size=dx, axis=0, accuracy=3)
        npt.assert_allclose(dFdX, jnp.cos(X) * jnp.cos(Y), rtol=1e-3)

        dFdY = difference(F, step_size=dy, axis=1, accuracy=3)
        npt.assert_allclose(dFdY, -jnp.sin(X) * jnp.sin(Y), atol=1e-7)

        # 1 4 6 3 2
        # left = (4-1)-> forward difference
        # right = (4-1), (6-4), (3-6), (2-3) = -1 -> backward difference
        npt.assert_allclose(
            difference(jnp.array([1, 4, 6, 3, 2]), accuracy=1, axis=0, derivative=1, method="backward"),
            jnp.array([3., 3., 2., -3., -1.]),
        )

        # 1 4 6 3 2 4
        # Forward Coeffs -> [-1.5  2.  -0.5] (0, 1, 2)
        # Backward Coeffs -> [ 0.5 -2.   1.5] (-2, -1, 0)
        npt.assert_allclose(
            difference(jnp.array([1, 4, 6, 3, 2, 4]), accuracy=2, axis=0, derivative=1, method="backward"),
            jnp.array([3.5, 4.5, 1.5, -5.5, 0.0, 3.5]),
        )

        # no central difference possible
        x = jnp.array([1.2, 1.4])
        # Forward Coeffs -> [-1 1] (0, 1)
        # Backward Coeffs -> [1 -1] (-1, 0)
        npt.assert_allclose(
            difference(x, accuracy=1, axis=0, derivative=1, method="backward"), jnp.array([0.2, 0.2])
        )


@pytest.mark.parametrize(
        "method", ("central", "forward","backward")
)
def test_divergence(method):
    with enable_x64():
        x, y = [jnp.linspace(0, 1, 100)] * 2
        dx, dy = x[1] - x[0], y[1] - y[0]
        X, Y = jnp.meshgrid(x, y, indexing="ij")
        F1 = X**2 + Y**3
        F2 = X**4 + Y**3
        F = jnp.stack([F1, F2], axis=0)  # 2D vector field F = (F1, F2)
        divZ = divergence(F, step_size=(dx, dy), accuracy=7, method=method, keepdims=False)
        divZ_true = 2 * X + 3 * Y**2  # (dF1/dx) + (dF2/dy)
        npt.assert_allclose(divZ, divZ_true, atol=5e-7)

        divZ = divergence(F, step_size=(dx, dy), accuracy=7,  method=method, keepdims=True)
        divZ_true = 2 * X + 3 * Y**2  # (dF1/dx) + (dF2/dy)
        npt.assert_allclose(divZ, divZ_true[None], atol=5e-7)


@pytest.mark.parametrize(
        "method", ("central", "forward","backward")
)
def test_gradient(method):
    with enable_x64():
        x, y = [jnp.linspace(0, 1, 100)] * 2
        dx, dy = x[1] - x[0], y[1] - y[0]
        X, Y = jnp.meshgrid(x, y, indexing="ij")
        F = X**2 + Y**3
        gradF = gradient(F, step_size=(dx, dy), accuracy=7, method=method)
        gradF_true = jnp.stack([2 * X, 3 * Y**2], axis=0)
        npt.assert_allclose(gradF, gradF_true, atol=5e-7)


@pytest.mark.parametrize(
        "method,atol", [
            ("central", 1e-7),
            ("forward", 1e-6),
            ("backward",1e-7)
        ]
)
def test_laplacian(method, atol):
    with enable_x64():
        # needs float64 for higher accuracy
        x, y = [jnp.linspace(0, 1, 100)] * 2
        dx, dy = x[1] - x[0], y[1] - y[0]
        X, Y = jnp.meshgrid(x, y, indexing="ij")
        Z = X**4 + Y**3
        laplacianZ = laplacian(Z, step_size=(dx, dy), accuracy=10, method=method)
        laplacianZ_true = 12 * X**2 + 6 * Y
        npt.assert_allclose(laplacianZ, laplacianZ_true, atol=atol)


@pytest.mark.parametrize(
        "method", ("central", "forward","backward")
)
def test_curl(method):
    with enable_x64():
        x, y, z = [jnp.linspace(0, 1, 100)] * 3
        dx, dy, dz = x[1] - x[0], y[1] - y[0], z[1] - z[0]
        X, Y, Z = jnp.meshgrid(x, y, z, indexing="ij")
        F1 = X**2 + Y**2
        F2 = X**4 + Y**3
        F3 = 0 * Z
        F = jnp.stack([F1, F2, F3], axis=0)
        curl_Z = curl(F, step_size=(dx, dy, dz), accuracy=5, method=method)

        F1 = 0 * X
        F2 = 0 * Y
        F3 = 4 * X**3 - 2 * Y
        curl_Z_true = jnp.stack([F1, F2, F3], axis=0)

        npt.assert_allclose(curl_Z, curl_Z_true, atol=1e-7)

        x, y = [jnp.linspace(-1, 1, 50)] * 2
        dx, dy = x[1] - x[0], y[1] - y[0]
        X, Y = jnp.meshgrid(x, y, indexing="ij")
        F1 = jnp.sin(Y)
        F2 = jnp.cos(X)
        F = jnp.stack([F1, F2], axis=0)
        res = curl(F, accuracy=4, step_size=dx, method=method, keepdims=False)
        npt.assert_allclose(res, -jnp.sin(X) - jnp.cos(Y), atol=1e-4)

        res = curl(F, accuracy=4, step_size=dx, method=method, keepdims=True)
        npt.assert_allclose(res[0], -jnp.sin(X) - jnp.cos(Y), atol=1e-4)

    with pytest.raises(ValueError):
        curl(F[None], accuracy=4, step_size=dx, method=method, keepdims=True)


@pytest.mark.parametrize(
        "method", ("central", "forward","backward")
)
def test_jacobian(method):
    with enable_x64():
        x, y = [jnp.linspace(-1, 1, 100)] * 2
        dx, dy = x[1] - x[0], y[1] - y[0]
        X, Y = jnp.meshgrid(x, y, indexing="ij")
        F1 = X**2 * Y
        F2 = 5 * X + jnp.sin(Y)
        F = jnp.stack([F1, F2], axis=0)
        JF = jacobian(F, accuracy=4, method=method, step_size=(dx, dy))
        JF_true = jnp.array([[2 * X * Y, X**2], [5 * jnp.ones_like(X), jnp.cos(Y)]])
        npt.assert_allclose(JF, JF_true, atol=1e-7)

        x1, x2, x3 = [jnp.linspace(-1, 1, 100)] * 3
        dx1, dx2, dx3 = x1[1] - x1[0], x2[1] - x2[0], x3[1] - x3[0]
        X1, X2, X3 = jnp.meshgrid(x1, x2, x3, indexing="ij")
        Y1 = X1
        Y2 = 5 * X3
        Y3 = 4 * X2**2 - 2 * X3
        Y4 = X3 * jnp.sin(X1)
        Y = jnp.stack([Y1, Y2, Y3, Y4], axis=0)
        JY = jacobian(Y, accuracy=5, step_size=(dx1, dx2, dx3), method=method)

        JY_true = jnp.array(
            [
                [jnp.ones_like(X1), jnp.zeros_like(X1), jnp.zeros_like(X1)],
                [jnp.zeros_like(X1), jnp.zeros_like(X1), 5 * jnp.ones_like(X1)],
                [jnp.zeros_like(X1), 8 * X2, -2 * jnp.ones_like(X1)],
                [X3 * jnp.cos(X1), jnp.zeros_like(X1), jnp.sin(X1)],
            ]
        )

        npt.assert_allclose(JY, JY_true, rtol=1e-3, atol=1e-5)


@pytest.mark.parametrize(
        "method", ("central", "forward","backward")
)
def test_hessian(method):
    with enable_x64():
        x, y = [jnp.linspace(-1, 1, 100)] * 2
        dx, dy = x[1] - x[0], y[1] - y[0]
        X, Y = jnp.meshgrid(x, y, indexing="ij")

        F = X**2 * Y
        H = hessian(F, accuracy=4, step_size=(dx, dy), method=method)
        H_true = jnp.array([[2 * Y, 2 * X], [2 * X, jnp.zeros_like(X)]])
        npt.assert_allclose(H, H_true, atol=1e-7)

        F = X**3 + Y**3
        H = hessian(F, accuracy=4, step_size=(dx, dy), method=method)
        H_true = jnp.array([[6 * X, jnp.zeros_like(X)], [jnp.zeros_like(X), 6 * Y]])
        npt.assert_allclose(H, H_true, atol=1e-7)

        F = jnp.sin(X) + jnp.cos(Y)
        H = hessian(F, accuracy=4, step_size=(dx, dy), method=method)
        H_true = jnp.array(
            [[-jnp.sin(X), jnp.zeros_like(X)], [jnp.zeros_like(X), -jnp.cos(Y)]]
        )
        npt.assert_allclose(H, H_true, atol=1e-7)

        # x, y, z = [jnp.linspace(0, 0.5, 100)] * 3
        # dx, dy, dz = x[1] - x[0], y[1] - y[0], z[1] - z[0]
        # X, Y, Z = jnp.meshgrid(x, y, z, indexing="ij")
        # F = jnp.sin(X * Y * Z)
        # H = hessian(F, accuracy=5, step_size=(dx, dy, dz))
        # H_true = jnp.array(
        #     [
        #         [
        #             -jnp.sin(X * Y * Z) * Y**2 * Z**2,
        #             -jnp.sin(X * Y * Z) * X * Y * Z**2 + Z * jnp.cos(X * Y * Z),
        #             -jnp.sin(X * Y * Z) * X * Y**2 * Z + Y * jnp.cos(X * Y * Z),
        #         ],
        #         [
        #             -jnp.sin(X * Y * Z) * X * Y * Z**2 + Z * jnp.cos(X * Y * Z),
        #             -jnp.sin(X * Y * Z) * X**2 * Z**2,
        #             -jnp.sin(X * Y * Z) * X**2 * Y * Z + X * jnp.cos(X * Y * Z),
        #         ],
        #         [
        #             -jnp.sin(X * Y * Z) * X * Y**2 * Z + Y * jnp.cos(X * Y * Z),
        #             -jnp.sin(X * Y * Z) * X**2 * Y * Z + X * jnp.cos(X * Y * Z),
        #             -jnp.sin(X * Y * Z) * X**2 * Y**2,
        #         ],
        #     ]
        # )

        # npt.assert_allclose(H, H_true, atol=1e-7)


def test_generate_coeffs():
    with pytest.raises(ValueError):
        _generate_backward_offsets(derivative=0, accuracy=1)
    with pytest.raises(ValueError):
        _generate_backward_offsets(derivative=1, accuracy=0)

    with pytest.raises(ValueError):
        _generate_forward_offsets(derivative=0, accuracy=1)
    with pytest.raises(ValueError):
        _generate_forward_offsets(derivative=1, accuracy=0)

    with pytest.raises(ValueError):
        _generate_central_offsets(derivative=0, accuracy=1)
    with pytest.raises(ValueError):
        _generate_central_offsets(derivative=1, accuracy=0)

    with pytest.raises(ValueError):
        generate_finitediff_coeffs(offsets=[1, 2], derivative=2)


def test_difference_error():
    x = jnp.ones([2])
    with pytest.raises(ValueError):
        difference(x, accuracy=2)



# credits to Mahmoud Asem 2022 @KAIST
# functions that operate on arrays
# higher accuracy finite difference gradient might require
# setting jax.config.update("jax_enable_x64", True)

from __future__ import annotations

import functools as ft
from itertools import combinations

import jax
import jax.numpy as jnp

from finitediffx._src.utils import (
    _check_and_return,
    _generate_backward_offsets,
    _generate_central_offsets,
    _generate_forward_offsets,
    generate_finitediff_coeffs,
)

__all__ = (
    "hessian",
    "difference",
    "gradient",
    "jacobian",
    "laplacian",
    "curl",
    "divergence",
)


def _central_difference(
    x: jax.Array,
    *,
    axis: int,
    left_coeffs,
    center_coeffs,
    right_coeffs,
    left_offsets,
    center_offsets,
    right_offsets,
):
    size = x.shape[axis]
    sliced = ft.partial(jax.lax.slice_in_dim, x, axis=axis)

    # use central difference for interior points
    left_x = sum(
        coeff * sliced(offset, offset - center_offsets[0])
        for offset, coeff in zip(left_offsets, left_coeffs)
    )

    right_x = sum(
        coeff * sliced(size + (offset - center_offsets[-1]), size + (offset))
        for offset, coeff in zip(right_offsets, right_coeffs)
    )

    center_x = sum(
        coeff * sliced(offset - center_offsets[0], size + (offset - center_offsets[-1]))
        for offset, coeff in zip(center_offsets, center_coeffs)
    )

    return jnp.concatenate([left_x, center_x, right_x], axis=axis)


def _central_difference_edge(
    x,
    *,
    axis,
    left_coeffs,
    right_coeffs,
    left_offsets,
    right_offsets,
):
    size = x.shape[axis]
    sliced = ft.partial(jax.lax.slice_in_dim, x, axis=axis)

    left_x = sum(
        coeff * sliced(offset, offset + left_offsets[-1] + size % 2)
        for offset, coeff in zip(left_offsets, left_coeffs)
    )

    right_x = sum(
        coeff * sliced(size + (offset - left_offsets[-1]), size + (offset))
        for offset, coeff in zip(right_offsets, right_coeffs)
    )

    return jnp.concatenate([left_x, right_x], axis=axis)


def _forward_difference(
    x,
    *,
    axis,
    left_coeffs,
    right_coeffs,
    left_offsets,
    right_offsets,
):
    size = x.shape[axis]
    sliced = ft.partial(jax.lax.slice_in_dim, x, axis=axis)

    left_x = sum(
        coeff * sliced(offset - left_offsets[0], size + (offset - left_offsets[-1]))
        for offset, coeff in zip(left_offsets, left_coeffs)
    )

    right_x = sum(
        coeff * sliced(size + (offset - left_offsets[-1]), size + (offset))
        for offset, coeff in zip(right_offsets, right_coeffs)
    )

    return jnp.concatenate([left_x, right_x], axis=axis)


def _backward_difference(
    x,
    *,
    axis,
    left_coeffs,
    right_coeffs,
    left_offsets,
    right_offsets,
):
    size = x.shape[axis]
    sliced = ft.partial(jax.lax.slice_in_dim, x, axis=axis)

    left_x = sum(
        coeff * sliced(offset, offset - right_offsets[0])
        for offset, coeff in zip(left_offsets, left_coeffs)
    )

    right_x = sum(
        coeff * sliced(offset - right_offsets[0], size + (offset - right_offsets[-1]))
        for offset, coeff in zip(right_offsets, right_coeffs)
    )

    return jnp.concatenate([left_x, right_x], axis=axis)


@ft.partial(jax.jit, static_argnames=("accuracy", "axis", "derivative", "method"))
def difference(
    array: jax.Array,
    *,
    axis: int = 0,
    accuracy: int = 1,
    step_size: float | jax.Array = 1,
    derivative: int = 1,
    method: str = "central",
) -> jax.Array:
    """Compute the finite difference derivative along a given axis with a given accuracy
    using central difference for interior points and forward/backward difference for boundary points
    Similar to np.gradient, but with the option to specify accuracy, derivative and step size
    See: https://github.com/google/jax/blob/main/jax/_src/numpy/lax_numpy.py

    Args:
        x: input array
        axis: axis along which to compute the gradient. Default is 0
        accuracy: accuracy order of the gradient. Default is 1
        step_size: step size. Default is 1
        derivative: derivative order of the gradient. Default is 1
        method: the method to use (forward, central, backward). Default is central
    Returns:
        Finite difference derivative along the given axis

    Example:
        # dydx of a 2D array
        >>> x, y = [jnp.linspace(0, 1, 100)] * 2
        >>> dx, dy = x[1] - x[0], y[1] - y[0]
        >>> X, Y = jnp.meshgrid(x, y, indexing="ij")
        >>> F =  jnp.sin(X) * jnp.cos(Y)
        >>> dFdX = difference(F, step_size=dx, axis=0, accuracy=3, method="central")
        >>> dFdXdY = difference(dFdX, step_size=dy, axis=1, accuracy=3, method="central")

        # 1d finite difference derivative
        >>> x = jnp.array([1.2, 1.3, 2.2, 3., 4.5, 5.5, 6., 7., 8., 20.])
        >>> difference(x, accuracy=1)
        [ 0.0999999  0.5        0.85       1.15       1.25       0.75 0.75       1.         6.5       12.       ]


        # apply forward difference to the first element with accuracy 1
        x_1 = 1.3-1.2 = 0.1

        # apply central difference to interior elements with accuracy 2
        x_2 = (2.2-1.2)/2 = 0.5
        x_3 = (3.-1.3)/2 = 0.85
        x_4 = (4.5-2.2)/2 = 1.15
        x_5 = (5.5-3.)/2 = 1.25
        x_6 = (6.-4.5)/2 = 0.75
        x_7 = (7.-5.5)/2 = 0.75
        x_8 = (8.-6.)/2 = 1.
        x_9 = (20.-7.)/2 = 6.5

        # apply backward difference to the last element with accuracy 1
        x_10 = 20.-8. = 12.
    """
    size = array.shape[axis]

    center_offsets = _generate_central_offsets(derivative, accuracy + 1)
    center_coeffs = generate_finitediff_coeffs(center_offsets, derivative)

    left_offsets = _generate_forward_offsets(derivative, accuracy)
    left_coeffs = generate_finitediff_coeffs(left_offsets, derivative)

    right_offsets = _generate_backward_offsets(derivative, accuracy)
    right_coeffs = generate_finitediff_coeffs(right_offsets, derivative)

    # use central difference for interior points and
    # forward/backward difference for boundary points
    if method == "central" and size >= len(center_offsets):
        return _central_difference(
            array,
            axis=axis,
            left_coeffs=left_coeffs,
            center_coeffs=center_coeffs,
            right_coeffs=right_coeffs,
            left_offsets=left_offsets,
            center_offsets=center_offsets,
            right_offsets=right_offsets,
        ) / (step_size**derivative)

    if method == "central" and size >= len(left_offsets):
        # if size < len(center_offsets), use forward/backward
        # difference for interior points
        # if size >= len(left_offsets):
        return _central_difference_edge(
            array,
            axis=axis,
            left_coeffs=left_coeffs,
            right_coeffs=right_coeffs,
            left_offsets=left_offsets,
            right_offsets=right_offsets,
        ) / (step_size**derivative)

    if method == "forward" and size >= len(left_offsets):
        return _forward_difference(
            array,
            axis=axis,
            left_coeffs=left_coeffs,
            right_coeffs=right_coeffs,
            left_offsets=left_offsets,
            right_offsets=right_offsets,
        ) / (step_size**derivative)

    if method == "backward" and size >= len(right_offsets):
        return _backward_difference(
            array,
            axis=axis,
            left_coeffs=left_coeffs,
            right_coeffs=right_coeffs,
            left_offsets=left_offsets,
            right_offsets=right_offsets,
        ) / (step_size**derivative)

    msg = f"Size of the array along axis {axis} is smaller than the number of points required"
    msg += f" for the accuracy {accuracy} and derivative {derivative} requested"
    msg += f".\nSize must be larger than {len(left_offsets)}, but got {size}"
    msg += f".\nMethod should be 'central', 'forward', 'backward', but got {method}"
    raise ValueError(msg)


@ft.partial(jax.jit, static_argnames=("accuracy", "method"))
def gradient(
    array: jax.Array,
    *,
    accuracy: int | tuple[int, ...] = 1,
    step_size: float | tuple[float, ...] | jax.Array = 1,
    method: str = "central",
) -> jax.Array:
    """Compute the ∇F of input array where F is a scalar function of x and
    returns vectors of the same shape as x stacked along the first axis.

    Args:
        x: input array
        accuracy: accuracy order of the gradient. Default is 1, can be a tuple for each axis
        step_size: step size. Default is 1, can be a tuple for each axis
        method: the method to use (forward, central, backward). Default is central

    Index notation : dF/dxi

    Example:
        # ∇F of a 2D array
        >>> x, y = [jnp.linspace(0, 1, 100)] * 2
        >>> dx, dy = x[1] - x[0], y[1] - y[0]
        >>> X, Y = jnp.meshgrid(x, y, indexing="ij")
        >>> Z = X**2 + Y**3
        >>> dZdX , dZdY = gradient(Z, step_size=(dx,dy))
        >>> dZdx_true, dZdy_true= 2*X , 3*Y**2
        >>> numpy.testing.assert_allclose(dZdx, dZdx_true, atol=1e-4)
        >>> numpy.testing.assert_allclose(dZdy, dZdy_true, atol=1e-4)
    """
    accuracy = _check_and_return(accuracy, array.ndim, "accuracy")
    step_size = _check_and_return(step_size, array.ndim, "step_size")

    return jnp.stack(
        [
            difference(
                array,
                accuracy=acc,
                step_size=step,
                derivative=1,
                axis=axis,
                method=method,
            )
            for axis, (acc, step) in enumerate(zip(accuracy, step_size))
        ],
        axis=0,
    )


@ft.partial(jax.jit, static_argnames=("accuracy", "method"))
def jacobian(
    array: jax.Array,
    *,
    accuracy: int | tuple[int, ...] = 1,
    step_size: float | tuple[float, ...] | jax.Array = 1,
    method: str = "central",
) -> jax.Array:
    """Compute the ∂Fi/∂xj of input array where F is a vector function of x and
    returns vectors of the same shape as x stacked along the first axis.

    Args:
        x: input array
        accuracy: accuracy order of the gradient. Default is 1, can be a tuple for each axis
        step_size: step size. Default is 1, can be a tuple for each axis
        method: the method to use (forward, central, backward). Default is central

    Index notation: ∂Fi/∂xj

    Example:
        # F: R^2 -> R^2
        # F = [ x^2*y, 5x+siny ]
        # JF = [ [2xy, x^2], [5, cos(y)] ]
        >>> with jax.experimental.enable_x64():
        ...    x, y = [jnp.linspace(-1, 1, 100)] * 2
        ...    dx, dy = x[1] - x[0], y[1] - y[0]
        ...    X, Y = jnp.meshgrid(x, y, indexing="ij")
        ...    F1 = X**2 * Y
        ...    F2 = 5 * X + jnp.sin(Y)
        ...    F = jnp.stack([F1, F2], axis=0)
        ...    JF = jacobian(F, accuracy=4, step_size=(dx, dy))
        ...    JF_true = jnp.array([[2 * X * Y, X**2], [5*jnp.ones_like(X), jnp.cos(Y)]])
        ...    npt.assert_allclose(JF, JF_true, atol=1e-7)
    """
    accuracy = _check_and_return(accuracy, array.ndim - 1, "accuracy")
    step_size = _check_and_return(step_size, array.ndim - 1, "step_size")

    return jnp.stack(
        [
            gradient(xi, accuracy=accuracy, step_size=step_size, method=method)
            for xi in array
        ],
        axis=0,
    )


@ft.partial(jax.jit, static_argnames=("accuracy", "method", "keepdims"))
def divergence(
    array: jax.Array,
    *,
    accuracy: int | tuple[int, ...] = 1,
    step_size: float | tuple[float, ...] = 1,
    keepdims: bool = True,
    method: str = "central",
) -> jax.Array:
    """Compute the ∇.F of input array where F is a vector field whose components are the first axis of x
    and returns a scalar field

    Args:
        x: input array where the leading axis is the dimension of the vector field
        accuracy: accuracy order of the gradient. Default is 1, can be a tuple for each axis
        step_size: step size. Default is 1, can be a tuple for each axis
        method: the method to use (forward, central, backward). Default is central
        keepdims: whether to keep the leading dimension. Default is True.

    Index notation: dFi/dxi

    Example:
        # ∇.F of a 2D array
        >>> x, y = [jnp.linspace(0, 1, 100)] * 2
        >>> dx, dy = x[1] - x[0], y[1] - y[0]
        >>> X, Y = jnp.meshgrid(x, y, indexing="ij")
        >>> F1 = X**2 + Y**3
        >>> F2 = X**4 + Y**3
        >>> F = jnp.stack([F1, F2], axis=0) # 2D vector field F = (F1, F2)
        >>> divZ = divergence(F,step_size=(dx,dy), accuracy=7, keepdims=False)
        >>> divZ_true = 2*X + 3*Y**2  # (dF1/dx) + (dF2/dy)
        >>> numpy.testing.assert_allclose(divZ, divZ_true, atol=5e-4)
    """
    accuracy = _check_and_return(accuracy, array.ndim - 1, "accuracy")
    step_size = _check_and_return(step_size, array.ndim - 1, "step_size")

    result = sum(
        difference(
            array[axis],
            accuracy=acc,
            step_size=step,
            derivative=1,
            axis=axis,
            method=method,
        )
        for axis, (acc, step) in enumerate(zip(accuracy, step_size))
    )

    if keepdims:
        return jnp.expand_dims(result, axis=0)
    return result


@ft.partial(jax.jit, static_argnames=("accuracy", "method"))
def hessian(
    array: jax.Array,
    *,
    accuracy: int | tuple[int, ...] = 2,
    step_size: float | tuple[float, ...] | jax.Array = 1,
    method: str = "central",
) -> jax.Array:
    """Compute hessian of F: R^m -> R

    Args:
        x: input array
        accuracy: accuracy order of the gradient. Default is 2, can be a tuple for each axis
        method: the method to use (forward, central, backward). Default is central
        step_size: step size. Default is 1, can be a tuple for each axis

    Index notation: d2F/dxij

    Example:
        >>> x, y = [jnp.linspace(-1, 1, 100)] * 2
        >>> dx, dy = x[1] - x[0], y[1] - y[0]
        >>> X, Y = jnp.meshgrid(x, y, indexing="ij")

        >>> F = X**2 * Y
        >>> H = hessian(F, accuracy=4, step_size=(dx, dy))
        >>> H_true = jnp.array([[2 * Y, 2 * X], [2 * X, jnp.zeros_like(X)]])
        >>> npt.assert_allclose(H, H_true, atol=1e-7)
    """
    accuracy = _check_and_return(accuracy, array.ndim, "accuracy")
    step_size = _check_and_return(step_size, array.ndim, "step_size")
    axes = tuple(range(array.ndim))
    F = dict()

    # diag
    for axis in range(array.ndim):
        F[2 * axis] = difference(
            array,
            accuracy=accuracy[axis],
            step_size=step_size[axis],
            axis=axis,
            derivative=2,
            method=method,
        )

    # off-diag
    for ax1, ax2 in combinations(axes, 2):
        F[ax1 + ax2] = difference(
            difference(
                array,
                accuracy=accuracy[ax1],
                step_size=step_size[ax1],
                method=method,
                axis=ax1,
            ),
            accuracy=accuracy[ax2],
            step_size=step_size[ax2],
            method=method,
            axis=ax2,
        )

    return jnp.stack(
        [
            jnp.stack([F[ax1 + ax2] for ax2 in range(array.ndim)])
            for ax1 in range(array.ndim)
        ]
    )


@ft.partial(jax.jit, static_argnames=("accuracy", "method"))
def laplacian(
    array: jax.Array,
    *,
    accuracy: int | tuple[int, ...] = 1,
    step_size: float | tuple[float, ...] | jax.Array = 1,
    method: str = "central",
) -> jax.Array:
    """Compute the ΔF of input array.
    Args:
        x: input array
        accuracy: accuracy order of the gradient. Default is 1, can be a tuple for each axis
        step_size: step size. Default is 1, can be a tuple for each axis
        method: the method to use (forward, central, backward). Default is central

    Index notation: d2F/dxi2
    Example:
        # ΔF array
        >>> x, y = [jnp.linspace(0, 1, 100)] * 2
        >>> dx, dy = x[1] - x[0], y[1] - y[0]
        >>> X, Y = jnp.meshgrid(x, y, indexing="ij")
        >>> Z = X**4 + Y**3
        >>> laplacianZ = laplacian(Z, step_size=(dx,dy))
        >>> laplacianZ_true = 12*X**2 + 6*Y
        >>> numpy.testing.assert_allclose(laplacianZ, laplacianZ_true, atol=1e-4)
    """
    accuracy = _check_and_return(accuracy, array.ndim, "accuracy")
    step_size = _check_and_return(step_size, array.ndim, "step_size")

    return sum(
        difference(
            array, accuracy=acc, step_size=step, derivative=2, method=method, axis=axis
        )
        for axis, (acc, step) in enumerate(zip(accuracy, step_size))
    )


def _curl_2d(
    array: jax.Array,
    *,
    accuracy: int | tuple[int, ...] = 1,
    step_size: float | tuple[float, ...] | jax.Array = 1,
    method: str = "central",
    keepdims: bool = True,
) -> jax.Array:
    """Compute the ∇×F of input array where F is a vector field whose components are the first axis of x
    and returns a vector field

    Index notation: εijk dFk/dxj

    Args:
        x: input array where the leading axis is the dimension of the vector field
        accuracy: accuracy order of the gradient. Default is 1, can be a tuple for each axis
        step_size: step size. Default is 1, can be a tuple for each axis
        method: the method to use (forward, central, backward). Default is central
        keepdims: whether to keep the leading dimension of the vector field

    Example:
        >>> x,y = [jnp.linspace(-1,1,50)]*2
        >>> dx,dy = x[1]-x[0],y[1]-y[0]
        >>> X,Y = jnp.meshgrid(x,y, indexing="ij")
        >>> F1 = jnp.sin(Y)
        >>> F2 = jnp.cos(X)
        >>> F = jnp.stack([F1,F2], axis=0)
        >>> curl = curl_2d(F, accuracy=4, step_size=dx)
    """
    dF1dY = difference(
        array[0], accuracy=accuracy[1], step_size=step_size[1], method=method, axis=1
    )
    dF2dX = difference(
        array[1], accuracy=accuracy[0], step_size=step_size[0], method=method, axis=0
    )

    result = dF2dX - dF1dY

    if keepdims:
        return jnp.expand_dims(result, axis=0)

    return result


def _curl_3d(
    array: jax.Array,
    *,
    accuracy: int | tuple[int, ...] = 1,
    step_size: float | tuple[float, ...] | jnp.ndarry = 1,
    method: str = "central",
) -> jax.Array:
    dF1dY = difference(
        array[0], accuracy=accuracy[1], step_size=step_size[1], method=method, axis=1
    )
    dF1dZ = difference(
        array[0], accuracy=accuracy[2], step_size=step_size[2], method=method, axis=2
    )

    dF2dX = difference(
        array[1], accuracy=accuracy[0], step_size=step_size[0], method=method, axis=0
    )
    dF2dZ = difference(
        array[1], accuracy=accuracy[2], step_size=step_size[2], method=method, axis=2
    )

    dF3dX = difference(
        array[2], accuracy=accuracy[0], step_size=step_size[0], method=method, axis=0
    )
    dF3dY = difference(
        array[2], accuracy=accuracy[1], step_size=step_size[1], method=method, axis=1
    )

    return jnp.stack(
        [
            dF3dY - dF2dZ,
            dF1dZ - dF3dX,
            dF2dX - dF1dY,
        ],
        axis=0,
    )


@ft.partial(jax.jit, static_argnames=("accuracy", "method", "keepdims"))
def curl(
    array: jax.Array,
    *,
    accuracy: int | tuple[int, ...] = 1,
    step_size: float | tuple[float, ...] | jax.Array = 1,
    method: str = "central",
    keepdims: bool = True,
) -> jax.Array:
    """Compute the ∇×F of input array where F is a vector field whose components are the first axis of x
    and returns a vector field

    Index notation: εijk dFk/dxj

    Args:
        x: input array where the leading axis is the dimension of the vector field
        accuracy: accuracy order of the gradient. Default is 1, can be a tuple for each axis
        step_size: step size. Default is 1, can be a tuple for each axis
        method: the method to use (forward, central, backward). Default is central
        keepdims: whether to keep the leading dimension of the vector field (only for 2D)

    Example:
        Curl for a 3D vector field is defined as
        F = (F1, F2, F3)
        ∇×F = (dF3/dy - dF2/dz, dF1/dz - dF3/dx, dF2/dx - dF1/dy)
        >>> jax.config.update("jax_enable_x64", True)
        >>> x,y,z = [jnp.linspace(0, 1, 100)] * 3
        >>> dx,dy,dz = x[1]-x[0], y[1]-y[0], z[1]-z[0]
        >>> X,Y,Z = jnp.meshgrid(x,y,z, indexing="ij")
        >>> F1 = X**2 + Y**3
        >>> F2 = X**4 + Y**3
        >>> F3 = jnp.zeros_like(F1)
        >>> F = jnp.stack([F1,F2,F3], axis=0)
        >>> curlF = finitediffx.curl(F, step_size=(dx,dy,dz),  accuracy=6)
        >>> curlF_exact = jnp.stack([F1*0,F1*0, 4*X**3 - 3*Y**2], axis=0)
        >>> npt.assert_allclose(curlF, curlF_exact, atol=1e-7)

        Curl of 2D vector field is defined as
        >>> x,y = [jnp.linspace(-1,1,50)]*2
        >>> dx,dy = x[1]-x[0],y[1]-y[0]
        >>> X,Y = jnp.meshgrid(x,y, indexing="ij")
        >>> F1 = jnp.sin(Y)
        >>> F2 = jnp.cos(X)
        >>> F = jnp.stack([F1,F2], axis=0)
        >>> curl = curl_2d(F, accuracy=4, step_size=dx)
    """

    accuracy = _check_and_return(accuracy, array.ndim - 1, "accuracy")
    step_size = _check_and_return(step_size, array.ndim - 1, "step_size")

    if array.ndim == 4 and array.shape[0] == 3:
        return _curl_3d(array, accuracy=accuracy, method=method, step_size=step_size)

    if array.ndim == 3 and array.shape[0] == 2:
        return _curl_2d(
            array,
            accuracy=accuracy,
            step_size=step_size,
            method=method,
            keepdims=keepdims,
        )

    msg = f"`curl` is only implemented for 2D and 3D vector fields, got {array.ndim}D"
    msg += "for 2D vector fields, the leading axis must have a shape=2, "
    msg += "for 3D vector fields, the leading axis must have a shape=3"
    raise ValueError(msg)
